unit GsGoMoku;

{

  TGsGoMoku
 ______________________________________________________________________

  Version:       1.03 (development)
  First version: (1.01) August, 1998
  Last change:   March, 2001
  Author:        Radu Buzea (radu@rospotline.com)

  Copyright  1998-2001 by Radu Buzea. All rights reserved.
 ______________________________________________________________________

  DESCRIPTION: 

  TGsGoMoku control is an implementation of the famous Go-Moku game.
  Go-Moku is an ancient game that has its origins in Japan. It is
  known as "Japanese tic-tac-toe" and is considered by many to be the
  world's greatest strategy game. "Go-Moku" which translates to "five
  stones" is played on the traditional Go board with Go stone pieces.
  The full name "Go-Moku Narabe" literally means "five stones in a
  row." In Japan, a more complex version of this game exists, known
  as Renju. Other name used for Go-Moku are: Wuziqi, Gomoku, Connect5,
  Five-in-a-Row, Five-in-Line, Rendzu, Go-bang, Gobang, Piskvorky,
  Noughts & Crosses, Cross-Zero, Omok, O-mok, Kakugo, Luffarschack,
  RistiNolla, Pente.

  Go-Moku is played on a 19 x 19 board. Players take turns placing
  their stones (black and white) on the board. The goal is to place
  five of your stones in a row of connected squares. The winning row
  may be horizontal, vertical, or diagonal. The user will start the
  game by clicking in any of the boxes and then the system will take
  its turn. The user can select any level (Easy or tough). There are
  fewer chances to win the game in the tough level. Each player will
  get five points for every winning game.

  The game is deceptively simple. The first five moves for example,
  can be made in 5,962,870,725,840 (that's five trillion) different
  ways.
 ______________________________________________________________________

  LICENSE AGREEMENT:

  Redistribution and use in source and binary form, with or without
  modification, are permitted provided if the following conditions
  are met:

  1. Redistribution of this source code must retain the above
  copyright notice, this license agreement and the following
  disclaimer (found at the end of this licence agreement). If you
  alter the source code in any way, however, then you have to submit
  these changes to the author. If you want to redistribute the source
  code with your changes, the complete original and unmodified source
  code must be distributed with the modified version.

  2. Redistribution of TGsTetris in source code or object form
  (including but not limited to .pas, .dcu, .obj, .res), in whole or
  in part, modified or unmodified, may not be redistributed for
  profit or as part of another commercial or shareware software
  package without express written permission from the author.

  3. Distribution in binary form must reproduce the above copyright
  notice, this license agreement and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
  Software using this code must contain a visible line of credit.
 ______________________________________________________________________

  DISCLAIMER:

  THIS SOFTWARE IS PROVIDED BY THE AUTHOR 'AS-IS', WITHOUT ANY EXPRESS
  OR IMPLIED WARRANTY. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
  IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  Radu Buzea (radu@rospotline.com)
 ______________________________________________________________________

}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls;

type
  TGsBevelStyle = (bvNone, bvRaised, bvLowered);
  TGsGoStone = (goNone, goPlayer1, goPlayer2);
  TGsGoPlayerSequence = (psPlayer1, psPlayer2);
  TGsGoPlayer = (plHuman, plComputer);
  TGsComputerSkill = (csLow, csMedium, csHigh);
  TOnClickSquare = procedure(Sender: TObject; SquareX, SquareY: Integer;
    CanChange: Boolean; var AllowChange: Boolean; var Stone: TGsGoStone) of Object;
  TOnDrawSquare = procedure(Sender: TObject; SquareX, SquareY: Integer;
    R: TRect; var Stone: TGsGoStone) of Object;
  TOnMouseOverSquare = procedure(Sender: TObject; SquareX, SquareY: Integer;
    Stone: TGsGoStone) of Object;
  TOnEndGame = procedure(Sender: TObject; WinnerStone: TGsGoStone) of Object;

  TGsGoMoku = class(TCustomControl)
  private
    FBorderStyle: TBorderStyle;
    FSquareDim: Integer;
    FBevelWidth: Integer;
    FNodes: Integer;
    FSquareBevelWidth: Integer;
    FMargin: Integer;
    FBGColor: TColor;
    FBGFiveColor: TColor;
    FMarginColor: TColor;
    FFGLightColor: TColor;
    FFGShadowColor: TColor;
    FBevel: TGsBevelStyle;
    FSquareBevel: TGsBevelStyle;
    FBuffer: TBitmap;
    FStone1: TBitmap;
    FStone2: TBitmap;
    FPlayer1: TGsGoPlayer;
    FPlayer2: TGsGoPlayer;
    FComputerSkill: TGsComputerSkill;
    FPlayerSequence: TGsGoPlayerSequence;
    FUseCursors: Boolean;
    FChessBrdStyle: Boolean;
    FAutoFGColors: Boolean;
    FLockUndo: Boolean;
    FUpdateBuffer: Boolean;
    FOwnerDraw: Boolean;
    FUndoList: TStrings;
    FRedoList: TStrings;
    FOnChange: TNotifyEvent;
    FOnClickSquare: TOnClickSquare;
    FOnDrawSquare: TOnDrawSquare;
    FOnMouseOverSquare: TOnMouseOverSquare;
    FOnPaint: TNotifyEvent;
    FOnEndGame: TOnEndGame;
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetBGColor(Value: TColor);
    procedure SetFGLightColor(Value: TColor);
    procedure SetFGShadowColor(Value: TColor);
    procedure SetSquareBevel(Value: TGsBevelStyle);
    procedure SetMargin(Value: Integer);
    procedure SetMarginColor(Value: TColor);
    procedure SetSquareBevelWidth(Value: Integer);
    procedure SetBevelWidth(Value: Integer);
    procedure SetBevel(Value: TGsBevelStyle);
    procedure SetStone1(Value: TBitmap);
    procedure SetStone2(Value: TBitmap);
    procedure SetChessBrdStyle(Value: Boolean);
    procedure SetAutoFGColors(Value: Boolean);
  protected
    FSaveSquareX: Integer;
    FSaveSquareY: Integer;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure Paint; override;
    procedure DrawStone(Support: TCanvas; X, Y: Integer; Is5InARow: Boolean);
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    function GetSquareAtPos(X, Y: Integer): TGsGoStone;
    procedure Click; override;
    procedure ChangeStone(X, Y: Integer; Stone: TGsGoStone);
    procedure ComputerSequence;
    procedure CalcBestPosition(Defensive: Boolean; NodeList: TStrings);
    function FiveInARow(Draw5InARow: Boolean): Boolean;
    function DefensiveTrickCheck(Stone: TGsGoStone; NodeList: TStrings): Boolean;
    function HorizontallyCheck(Defensive: Boolean; Stone: TGsGoStone; NodeList: TStrings): Boolean;
    function VerticallyCheck(Defensive: Boolean; Stone: TGsGoStone; NodeList: TStrings): Boolean;
    function ObliqueCheck_TLBR(Defensive: Boolean; Stone: TGsGoStone; NodeList: TStrings): Boolean;
    function ObliqueCheck_TRBL(Defensive: Boolean; Stone: TGsGoStone; NodeList: TStrings): Boolean;
  public
    GoMokuBoard: array[1..19, 1..19] of TGsGoStone;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure NewGame;
    procedure LoadFromFile(const FileName: TFileName);
    procedure SaveToFile(const FileName: TFileName);
    function CanUndo: Boolean;
    function CanRedo: Boolean;
    procedure Undo;
    procedure Redo;
  published
    property AutoForegroundColors: Boolean read FAutoFGColors write SetAutoFGColors default False;
    property Background: TColor read FBGColor write SetBGColor default clBtnFace;
    property Backround5InARow: TColor read FBGFiveColor write FBGFiveColor default clRed;
    property Bevel: TGsBevelStyle read FBevel write SetBevel default bvNone;
    property BevelWidth: Integer read FBevelWidth write SetBevelWidth default 1;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsSingle;
    property ChessBoardStyle: Boolean read FChessBrdStyle write SetChessBrdStyle default False;
    property ComputerSkill: TGsComputerSkill read FComputerSkill write FComputerSkill default csMedium;
    property DragCursor;
    property DragMode;
    property ForegroundLight: TColor read FFGLightColor write SetFGLightColor default clBtnHighlight;
    property ForegroundShadow : TColor read FFGShadowColor write SetFGShadowColor default clBtnShadow;
    property Hint;
    property LockUndo: Boolean read FLockUndo write FLockUndo default False;
    property Margin: Integer read FMargin write SetMargin default 6;
    property MarginColor: TColor read FMarginColor write SetMarginColor default clAppWorkspace;
    property OwnerDraw: Boolean read FOwnerDraw write FOwnerDraw default False;
    property ParentShowHint;
    property Player1: TGsGoPlayer read FPlayer1 write FPlayer1;
    property Player1Stone: TBitmap read FStone1 write SetStone1;
    property Player2: TGsGoPlayer read FPlayer2 write FPlayer2;
    property Player2Stone: TBitmap read FStone2 write SetStone2;
    property PlayerSequence: TGsGoPlayerSequence read FPlayerSequence write FPlayerSequence default psPlayer1;
    property ShowHint;
    property SquareBevel: TGsBevelStyle read FSquareBevel write SetSquareBevel default bvRaised;
    property SquareBevelWidth: Integer read FSquareBevelWidth write SetSquareBevelWidth default 1;
    property UseCursors: Boolean read FUseCursors write FUseCursors default True;
    property Visible;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnClick;
    property OnClickSquare: TOnClickSquare read FOnClickSquare write FOnClickSquare;
    property OnDblClick;
    property OnDrawSquare: TOnDrawSquare read FOnDrawSquare write FOnDrawSquare;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEndGame: TOnEndGame read FOnEndGame write FOnEndGame;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseOverSquare: TOnMouseOverSquare read FOnMouseOverSquare write FOnMouseOverSquare;
    property OnMouseUp;
    property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
    property OnStartDrag;
  end;

procedure Register;

const
  crCanClick = 1;
  crCannotClick = 2;
  SwitchSequence: array[TGsGoPlayerSequence] of TGsGoPlayerSequence = (psPlayer2, psPlayer1);

var
  FCursorClick1, FCursorClick2: TCursor;

implementation

{$R GSGOMOKU.RES}
{$R GSTETRIS_IDE.RES}

type
  TGsGoMokuFileRec = record
    Ver: array[0..16] of Char;
    Player1: Word;
    Player2: Word;
    Player1Turn: Boolean;
    Lines: array[1..19, 1..19] of Word;
  end;

  TGsGoMokuBrainNode = class(TObject)
  private
    FX, FY: Integer;
    FLevel: Integer;
    FStone: TGsGoStone;
  public
    property X: Integer read FX write FX;
    property Y: Integer read FY write FY;
    property Level: Integer read FLevel write FLevel;
    property Stone: TGsGoStone read FStone write FStone;
  end;

(* TGsGoMoku ****************************************************************************)

constructor TGsGoMoku.Create(AOwner: TComponent);
var
  X, Y: Integer;
begin
  inherited Create(AOwner);
  Width := 16;
  Height := 16;
  FMargin := 6;  
  for X := 1 to 19 do
    for Y := 1 to 19 do
      GoMokuBoard[X, Y] := goNone;
  FBuffer := TBitmap.Create;
  FStone1 := TBitmap.Create;
  FStone1.LoadFromResourceName(hInstance, 'WHITESTONE');
  FStone2 := TBitmap.Create;
  FStone2.LoadFromResourceName(hInstance, 'BLACKSTONE');
  FUndoList := TStringList.Create;
  FRedoList := TStringList.Create;
  ParentColor := False;
  FBGColor := clBtnFace;
  FBGFiveColor := clRed;
  FFGLightColor := clBtnHighlight;
  FFGShadowColor := clBtnShadow;
  FMarginColor := clAppWorkspace;
  Color := FMarginColor;
  FBorderStyle := bsSingle;
  FBevel := bvNone;
  FBevelWidth := 1;
  FSquareBevel := bvRaised;
  FSquareBevelWidth := 1;
  FAutoFGColors := True;
  FChessBrdStyle := True;
  FLockUndo := False;
  FUpdateBuffer := True;
  FUseCursors := True;
  FPlayer1 := plHuman;
  FPlayer2 := plHuman;
  FComputerSkill := csMedium;
  FNodes := 0;
end;

destructor TGsGoMoku.Destroy;
begin
  FBuffer.Free;
  FStone1.Free;
  FStone2.Free;
  FUndoList.Free;
  FRedoList.Free;
  inherited Destroy;
end;

procedure TGsGoMoku.SetBorderStyle(Value: TBorderStyle);
begin
  if FBorderStyle <> Value then
  begin
    FBorderStyle := Value;
    FUpdateBuffer := True;
    RecreateWnd;
  end;
end;

procedure TGsGoMoku.SetBGColor(Value: TColor);
begin
  if FBGColor <> Value then
  begin
    FBGColor := Value;
    FUpdateBuffer := True;
    Invalidate;
  end;
end;

procedure TGsGoMoku.SetFGLightColor(Value: TColor);
begin
  if FFGLightColor <> Value then
  begin
    FFGLightColor := Value;
    FUpdateBuffer := True;
    if not FAutoFGColors then Invalidate;
  end;
end;

procedure TGsGoMoku.SetFGShadowColor(Value: TColor);
begin
  if FFGShadowColor <> Value then
  begin
    FFGShadowColor := Value;
    FUpdateBuffer := True;
    if not FAutoFGColors then Invalidate;
  end;
end;

procedure TGsGoMoku.SetSquareBevel(Value: TGsBevelStyle);
begin
  if FSquareBevel <> Value then
  begin
    FSquareBevel := Value;
    FUpdateBuffer := True;
    Invalidate;
  end;
end;

procedure TGsGoMoku.SetMargin(Value: Integer);
begin
  if Value < 0 then Value := 0;
  if Value > 16 then Value := 16;
  if FMargin <> Value then
  begin
    FMargin := Value;
    if (FBevel <> bvNone) and (FMargin < FBevelWidth) then FMargin := FBevelWidth;
    FUpdateBuffer := True;
    RecreateWnd;
  end;
end;

procedure TGsGoMoku.SetMarginColor(Value: TColor);
begin
  if FMarginColor <> Value then
  begin
    FMarginColor := Value;
    Color := Value;
    FUpdateBuffer := True;
    Invalidate;
  end;
end;

procedure TGsGoMoku.SetSquareBevelWidth(Value: Integer);
begin
  if Value < 1 then Value := 1;
  if Value > 4 then Value := 4;
  if FSquareBevelWidth <> Value then
  begin
    FSquareBevelWidth := Value;
    FUpdateBuffer := True;
    Invalidate;
  end;
end;

procedure TGsGoMoku.SetBevelWidth(Value: Integer);
begin
  if Value < 1 then Value := 1;
  if Value > 6 then Value := 6;
  if FBevelWidth <> Value then
  begin
    FBevelWidth := Value;
    if (FBevel <> bvNone) and (FMargin < FBevelWidth) then FMargin := FBevelWidth;
    FUpdateBuffer := True;
    RecreateWnd;
  end;
end;

procedure TGsGoMoku.SetBevel(Value: TGsBevelStyle);
begin
  if FBevel <> Value then
  begin
    FBevel := Value;
    if (FBevel <> bvNone) and (FMargin < FBevelWidth) then FMargin := FBevelWidth;
    FUpdateBuffer := True;
    Invalidate;
  end;
end;

procedure TGsGoMoku.SetStone1(Value: TBitmap);
begin
  FStone1.Assign(Value);
  if not (csDesigning in ComponentState) then
  begin
    FUpdateBuffer := True;
    Invalidate;
  end;
end;

procedure TGsGoMoku.SetStone2(Value: TBitmap);
begin
  FStone2.Assign(Value);
  if not (csDesigning in ComponentState) then
  begin
    FUpdateBuffer := True;
    Invalidate;
  end;    
end;

procedure TGsGoMoku.SetChessBrdStyle(Value: Boolean);
begin
  if FChessBrdStyle <> Value then
  begin
    FChessBrdStyle := Value;
    FUpdateBuffer := True;
    Invalidate;
  end;
end;

procedure TGsGoMoku.SetAutoFGColors(Value: Boolean);
begin
  if FAutoFGColors <> Value then
  begin
    FAutoFGColors := Value;
    FUpdateBuffer := True;
    Invalidate;
  end;
end;

procedure TGsGoMoku.CreateParams(var Params: TCreateParams);
const
  BorderStyles: array [TBorderStyle] of DWord = (0, WS_BORDER);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or BorderStyles[FBorderStyle];
    if (NewStyleControls or Ctl3D) and (FBorderStyle = bsSingle) then
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
  end;
end;

procedure TGsGoMoku.CreateWnd;
begin
  inherited CreateWnd;
  if not (csDesigning in ComponentState) then
  begin
    FCursorClick1 := LoadCursor(hInstance, 'CANCLICK');
    Screen.Cursors[crCanClick] := FCursorClick1;
    FCursorClick2 := LoadCursor(hInstance, 'CANNOTCLICK');
    Screen.Cursors[crCannotClick] := FCursorClick2;
  end;  
end;

procedure TGsGoMoku.WMSize(var Message: TWMSize);
var
  CrDim: Integer;
begin
  inherited;
  FUpdateBuffer := True;
  CrDim := FMargin * 2;
  FSquareDim := (ClientWidth - CrDim) div 19;
  if FSquareDim < 18 then FSquareDim := 18; 
  CrDim := (FSquareDim * 19) + CrDim;
  if ClientWidth <> CrDim then
  begin
    ClientWidth := CrDim;
    Exit;
  end;
  if ClientHeight <> ClientWidth then
  begin
    ClientHeight := CrDim;
    Exit;
  end;
  FBuffer.Width := ClientWidth;
  FBuffer.Height := ClientHeight;
end;

procedure IncColor(var Color: TColor; Increment: Integer);
var
  R, G, B: Byte;
  LColor: LongInt;
begin
  LColor := ColorToRGB(Color);
  if Increment = 0 then Increment := 1;
  if Increment > 0 then
  begin
    if GetRValue(LColor) < 255 - Increment then
      R := GetRValue(LColor) + Increment
    else
      R := 255;
    if GetGValue(LColor) < 255 - Increment then
      G := GetGValue(LColor) + Increment
    else
      G := 255;
    if GetBValue(LColor) < 255 - Increment then
      B := GetBValue(LColor) + Increment
    else
      B := 255;
  end else
  begin
    if GetRValue(LColor) + Increment > 0  then
      R := GetRValue(LColor) + Increment
    else
      R := 0;
    if GetGValue(LColor) + Increment > 0 then
      G := GetGValue(LColor) + Increment
    else
      G := 0;
    if GetBValue(LColor) + Increment > 0 then
      B := GetBValue(LColor) + Increment
    else
      B := 0;
  end;
  Color := RGB(R, G, B);
end;

procedure TGsGoMoku.Paint;
var
  R: TRect;
  X, Y: Integer;
  LightCl, DrakCl: TColor;
begin
  if (csLoading in ComponentState) or (csDestroying in ComponentState) then
  begin
    inherited Paint;
    Exit;
  end;
  if Assigned(FOnPaint) then
  begin
    FOnPaint(Self);
    if FOwnerDraw then Exit;
  end;
  if FUpdateBuffer then
  begin
    R := ClientRect;
    FBuffer.Canvas.Brush.Color := FMarginColor;
    FBuffer.Canvas.FillRect(R);
    if not FAutoFGColors then
    begin
      LightCl := FFGLightColor;
      DrakCl := FFGShadowColor;
    end else
    begin
      LightCl := FMarginColor;
      IncColor(LightCl, 50);
      DrakCl := FMarginColor;
      IncColor(DrakCl, -50);
    end;
    case FBevel of
      bvRaised:  Frame3D(FBuffer.Canvas, R, LightCl, DrakCl, FBevelWidth);
      bvLowered: Frame3D(FBuffer.Canvas, R, DrakCl, LightCl, FBevelWidth);
    end;
    for X := 0 to 18 do
      for Y := 0 to 18 do
        DrawStone(FBuffer.Canvas, X + 1, Y  + 1, False);
    FUpdateBuffer := False;    
  end;
  Canvas.Draw(0, 0, FBuffer);
  FiveInARow(True);
end;

procedure TGsGoMoku.DrawStone(Support: TCanvas; X, Y: Integer;
  Is5InARow: Boolean);
var
  TmpBMP: TBitmap;
  CBrdColor, StoneBGColor: TColor;
  SquareR, R1, R2: TRect;
  L, T: Integer;
  Stone: TGsGoStone;
  R, G, B: Byte;
  LColor: LongInt;

  procedure DrawFinally;
  var
    LightCl, DrakCl: TColor;
  begin
    Support.Draw(L, T, TmpBMP);
    if not FAutoFGColors then
    begin
      LightCl := FFGLightColor;
      DrakCl := FFGShadowColor;
    end else
    begin
      LightCl := TmpBMP.Canvas.Brush.Color;
      IncColor(LightCl, 75);
      DrakCl := TmpBMP.Canvas.Brush.Color;
      IncColor(DrakCl, -75);
    end;
    case FSquareBevel of
      bvRaised:  Frame3D(Support, SquareR, LightCl, DrakCl, FSquareBevelWidth);
      bvLowered: Frame3D(Support, SquareR, DrakCl, LightCl, FSquareBevelWidth);
    end;
    TmpBMP.Free;
  end;

begin
  if not (X in [1..19]) or not (Y in [1..19]) then Exit;
  if Support = nil then Exit;
  Stone := GoMokuBoard[X, Y];
  L := ((X - 1) * FSquareDim) + FMargin;
  T := ((Y - 1) * FSquareDim) + FMargin;
  SquareR := Rect(L, T, L + FSquareDim, T + FSquareDim);
  if Assigned(FOnDrawSquare) then
  begin
    FOnDrawSquare(Self, X, Y, SquareR, Stone);
    if FOwnerDraw then Exit;
  end;
  TmpBMP := TBitmap.Create;
  TmpBMP.Width := FSquareDim;
  TmpBMP.Height := FSquareDim;
  if FChessBrdStyle then
  begin
    if Is5InARow then
      LColor := ColorToRGB(FBGFiveColor)
    else
      LColor := ColorToRGB(FBGColor);
    if GetRValue(LColor) < 230 then
      R := GetRValue(LColor) + 25
    else
      R := GetRValue(LColor) - 25;
    if GetGValue(LColor) < 230 then
      G := GetGValue(LColor) + 25
    else
      G := GetGValue(LColor) - 25;
    if GetBValue(LColor) < 230 then
      B := GetBValue(LColor) + 25
    else
      B := GetBValue(LColor) - 25;
    if Is5InARow then
    begin
      if FBGFiveColor = clBlack then
        CBrdColor := RGB(65, 65, 65)
      else
        CBrdColor := RGB(R, G, B);
    end else
      if FBGColor = clBlack then
        CBrdColor := RGB(65, 65, 65)
      else
        CBrdColor := RGB(R, G, B);
    if (X mod 2) > 0 then
    begin
      if (Y mod 2) > 0 then
        TmpBMP.Canvas.Brush.Color := CBrdColor
      else
        if Is5InARow then
          TmpBMP.Canvas.Brush.Color := FBGFiveColor
        else
          TmpBMP.Canvas.Brush.Color := FBGColor;
    end else
    begin
      if (Y mod 2) = 0 then
        TmpBMP.Canvas.Brush.Color := CBrdColor
      else
        if Is5InARow then
          TmpBMP.Canvas.Brush.Color := FBGFiveColor
        else
          TmpBMP.Canvas.Brush.Color := FBGColor;
    end;
  end else
    if Is5InARow then
      TmpBMP.Canvas.Brush.Color := FBGFiveColor
    else
      TmpBMP.Canvas.Brush.Color := FBGColor;
  R1 := Rect(0, 0, FSquareDim, FSquareDim);
  TmpBMP.Canvas.FillRect(R1);
  if csDesigning in ComponentState then
  begin
    DrawFinally;
    Exit;
  end;
  case Stone of
    goPlayer1: if not FStone1.Empty then
      begin
        StoneBGColor := FStone1.Canvas.Pixels[1, 1];
        R2 := Rect(0, 0, FStone1.Width, FStone1.Height);
        if FSquareDim > FStone1.Width then
        begin
          R1.Left := (FSquareDim - FStone1.Width) div 2;
          R1.Right := R1.Left + FStone1.Width;
        end;
        if FSquareDim > FStone1.Height then
        begin
          R1.Top := (FSquareDim - FStone1.Height) div 2;
          R1.Bottom := R1.Top + FStone1.Height;
        end;
        TmpBMP.Canvas.BrushCopy(R1, FStone1, R2, StoneBGColor);
      end;
    goPlayer2: if not FStone2.Empty then
      begin
        StoneBGColor := FStone2.Canvas.Pixels[1, 1];
        R2 := Rect(0, 0, FStone2.Width, FStone2.Height);
        if FSquareDim > FStone2.Width then
        begin
          R1.Left := (FSquareDim - FStone2.Width) div 2;
          R1.Right := R1.Left + FStone2.Width;
        end;
        if FSquareDim > FStone2.Height then
        begin
          R1.Top := (FSquareDim - FStone2.Height) div 2;
          R1.Bottom := R1.Top + FStone2.Height;
        end;
        TmpBMP.Canvas.BrushCopy(R1, FStone2, R2, StoneBGColor);
      end;
  end;
  DrawFinally;
end;

procedure TGsGoMoku.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if Button <> mbLeft then Exit;
  if (X > FMargin) and (Y > FMargin) and
   (X < ClientWidth - FMargin) and (Y < ClientHeight - FMargin) then
  begin
    FSaveSquareX := ((X - FMargin) div FSquareDim) + 1;
    FSaveSquareY := ((Y - FMargin) div FSquareDim) + 1;
    if FSaveSquareX < 1 then FSaveSquareX := 1;
    if FSaveSquareX > 19 then FSaveSquareX := 19;
    if FSaveSquareY < 1 then FSaveSquareY := 1;
    if FSaveSquareY > 19 then FSaveSquareY := 19;
  end else
  begin
    FSaveSquareX := 0;
    FSaveSquareX := 0;
  end;
end;

procedure TGsGoMoku.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  CrVal: Integer;
begin
  inherited MouseMove(Shift, X, Y);
  CrVal := Cursor;
  if not FiveInARow(False) and (GetSquareAtPos(X, Y) = goNone) then
  begin
    if (X > FMargin) and (Y > FMargin) and (X < ClientWidth - FMargin) and
      (Y < ClientHeight - FMargin) then
    begin
      if FUseCursors then CrVal := crCanClick;
    end else
      if FUseCursors then CrVal := crDefault;
  end else
    if FUseCursors then CrVal := crCannotClick;
  if Cursor <> CrVal then Cursor := CrVal;
end;

function TGsGoMoku.GetSquareAtPos(X, Y: Integer): TGsGoStone;
var
  SquareX, SquareY: Integer;
begin
  if (X > FMargin) and (Y > FMargin) and
   (X < ClientWidth - FMargin) and (Y < ClientHeight - FMargin) then
  begin
    SquareX := ((X - FMargin) div FSquareDim) + 1;
    SquareY := ((Y - FMargin) div FSquareDim) + 1;
    if SquareX < 1 then SquareX := 1;
    if SquareX > 19 then SquareX := 19;
    if SquareY < 1 then SquareY := 1;
    if SquareY > 19 then SquareY := 19;
    if Assigned(FOnMouseOverSquare) then
      FOnMouseOverSquare(Self, SquareX, SquareY, GoMokuBoard[SquareX, SquareY]);
    Result := GoMokuBoard[SquareX, SquareY];
  end else
    Result := goNone;
end;


procedure TGsGoMoku.Click;
const
  DoStone: array[TGsGoPlayerSequence] of TGsGoStone = (goPlayer1, goPlayer2);
var
  P: TGsGoStone;
  CanChange, AllowChange: Boolean;
begin
  if (FSaveSquareX <> 0) and (FSaveSquareY <> 0) and not FiveInARow(False) then
  begin
    P := GoMokuBoard[FSaveSquareX, FSaveSquareY];
    CanChange := (P = goNone);
    AllowChange := CanChange;
    if Assigned(FOnClickSquare) then
      FOnClickSquare(Self, FSaveSquareX, FSaveSquareY, CanChange, AllowChange, P);
    if CanChange then
    begin
      if AllowChange then
      begin
        ChangeStone(FSaveSquareX, FSaveSquareY, DoStone[FPlayerSequence]);
        FPlayerSequence := SwitchSequence[FPlayerSequence];
        FNodes := FNodes + 1;
        if Assigned(FOnChange) then FOnChange(Self);
        if FiveInARow(True) then Exit;
        if FNodes >= 361 then
        begin
          if Assigned(FOnEndGame) then FOnEndGame(Self, goNone);
          Exit;
        end;
        if (PlayerSequence = psPlayer1) and (FPlayer1 = plComputer) then
        begin
          ComputerSequence;
          Exit;
        end;
        if (PlayerSequence = psPlayer2) and (FPlayer2 = plComputer) then
        begin
          ComputerSequence;
          Exit;
        end;
      end else
        if (P <> GoMokuBoard[FSaveSquareX, FSaveSquareY]) then
          ChangeStone(FSaveSquareX, FSaveSquareY, P);
    end;
  end else
    inherited Click;
end;

procedure TGsGoMoku.ChangeStone(X, Y: Integer; Stone: TGsGoStone);
var
  uX, uY: Integer;
  S: String;
begin
  if not FLockUndo then
  begin
    S := '';
    for uY := 1 to 19 do
      for uX := 1 to 19 do
      case GoMokuBoard[uX, uY] of
        goPlayer1: S := S + '1';
        goPlayer2: S := S + '2';
      else
        S := S + '0';
      end;
    FUndoList.Add(S);
    FRedoList.Clear;
  end;  
  GoMokuBoard[X, Y] := Stone;
  DrawStone(Canvas, X, Y, False);
  DrawStone(FBuffer.Canvas, X, Y, False);
end;

procedure TGsGoMoku.NewGame;
var
  X, Y: Integer;
  SaveCursor: TCursor;
begin
  if FUseCursors then Cursor := crCanClick else Cursor := crDefault;
  SaveCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  FUndoList.Clear;
  FRedoList.Clear;
  FNodes := 0;
  for Y := 1 to 19 do
    for X := 1 to 19 do
    begin
      GoMokuBoard[X, Y] := goNone;
      DrawStone(Canvas, X, Y, False);
      DrawStone(FBuffer.Canvas, X, Y, False);
    end;
  Screen.Cursor := SaveCursor;
  if Assigned(FOnChange) then FOnChange(Self);
  if (PlayerSequence = psPlayer1) and (FPlayer1 = plComputer) then
  begin
    ComputerSequence;
    Exit;
  end;
  if (PlayerSequence = psPlayer2) and (FPlayer2 = plComputer) then
  begin
    ComputerSequence;
    Exit;
  end;
end;

procedure TGsGoMoku.LoadFromFile(const FileName: TFileName);
var
  GameFileRec: File of TGsGoMokuFileRec;
  OpenRec: TGsGoMokuFileRec;
  X, Y, FGSize: Integer;
  P: TGsGoStone;
  SaveCursor: TCursor;
begin
  if not FileExists(FileName) then Exit;
  if FUseCursors then Cursor := crCanClick else Cursor := crDefault;
  SaveCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  AssignFile(GameFileRec, FileName);
  Reset(GameFileRec);
  FGSize := FileSize(GameFileRec);
  if FGSize < 1 then
  begin
    CloseFile(GameFileRec);
    Screen.Cursor := SaveCursor;
    Exit;
  end;
  Seek(GameFileRec, 0);
  Read(GameFileRec, OpenRec);
  CloseFile(GameFileRec);
  for Y := 1 to 19 do
    for X := 1 to 19 do
    begin
      case OpenRec.Lines[X, Y] of
        1: P := goPlayer1;
        2: P := goPlayer2;
      else
        P := goNone;
      end;
      ChangeStone(X, Y, P);
    end;
  Screen.Cursor := SaveCursor;
  FUndoList.Clear;
  FRedoList.Clear;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TGsGoMoku.SaveToFile(const FileName: TFileName);
var
  GameFileRec: File of TGsGoMokuFileRec;
  SaveRec: TGsGoMokuFileRec;
  X, Y: Integer;
  W: Word;
begin
  AssignFile(GameFileRec, FileName);
  Rewrite(GameFileRec);
  SaveRec.Ver := 'GsGoMoku.File';
  if FPlayer1 = plHuman then SaveRec.Player1 := 1 else SaveRec.Player1 := 2;
  if FPlayer2 = plHuman then SaveRec.Player2 := 1 else SaveRec.Player2 := 2;
  SaveRec.Player1Turn := True;  
  for X := 1 to 19 do
    for Y := 1 to 19 do
    begin
      case GoMokuBoard[X, Y] of
        goPlayer1: W := 1;
        goPlayer2: W := 2;
      else
        W := 0;
      end;
      SaveRec.Lines[X, Y] := W;
    end;
  Write(GameFileRec, SaveRec);
  CloseFile(GameFileRec);
end;

function TGsGoMoku.CanUndo: Boolean;
begin
  Result := (FUndoList.Count > 0) and not FLockUndo;
end;

function TGsGoMoku.CanRedo: Boolean;
begin
  Result := (FRedoList.Count > 0) and not FLockUndo;
end;

procedure TGsGoMoku.Undo;
var
  S: string;
  I, X, Y: Integer;
  Stone : TGsGoStone;
  Has5InARow: Boolean;
begin
  if not CanUndo or FLockUndo then Exit;
  if FUseCursors then Cursor := crCanClick else Cursor := crDefault;
  Has5InARow := FiveInARow(False);
  S := '';
  for Y := 1 to 19 do
    for X := 1 to 19 do
    case GoMokuBoard[X, Y] of
      goPlayer1: S := S + '1';
      goPlayer2: S := S + '2';
    else
      S := S + '0';
    end;
  FRedoList.Add(S);
  S := FUndoList[pred(FUndoList.Count)];
  FUndoList.Delete(pred(FUndoList.Count));
  FPlayerSequence := SwitchSequence[FPlayerSequence];
  Y := 1;
  X := 1;
  for I := 1 to Length(S) do
  begin
    case S[I] of
      '1': Stone := goPlayer1;
      '2': Stone := goPlayer2;
    else
      Stone := goNone;
    end;
    if GoMokuBoard[X, Y] <> Stone then
    begin
      GoMokuBoard[X, Y] := Stone;
      DrawStone(FBuffer.Canvas, X, Y, False);
      if Has5InARow then
        Canvas.Draw(0, 0, FBuffer)
      else
        DrawStone(Canvas, X, Y, False);
      if FPlayer1 <> FPlayer2 then
        if ((FPlayerSequence = psPlayer1) and (FPlayer1 = plComputer)) or
           ((FPlayerSequence = psPlayer2) and (FPlayer2 = plComputer)) then
        begin
          Undo;
          Exit;
        end;
      if Assigned(FOnChange) then FOnChange(Self);
      Exit;
    end;
    X := X + 1;
    if X > 19 then
    begin
      X := 1;
      Y := Y + 1;
      if Y > 19 then Exit;
    end;
  end;
end;

procedure TGsGoMoku.Redo;
var
  S: string;
  I, X, Y: Integer;
  Stone : TGsGoStone;
begin
  if not CanRedo or FLockUndo then Exit;
  S := '';
  for Y := 1 to 19 do
    for X := 1 to 19 do
    case GoMokuBoard[X, Y] of
      goPlayer1: S := S + '1';
      goPlayer2: S := S + '2';
    else
      S := S + '0';
    end;
  FUndoList.Add(S);
  S := FUndoList[pred(FUndoList.Count)];
  S := FRedoList[pred(FRedoList.Count)];
  FRedoList.Delete(pred(FRedoList.Count));
  FPlayerSequence := SwitchSequence[FPlayerSequence];
  Y := 1;
  X := 1;
  for I := 1 to Length(S) do
  begin
    case S[I] of
      '1': Stone := goPlayer1;
      '2': Stone := goPlayer2;
    else
      Stone := goNone;
    end;
    if GoMokuBoard[X, Y] <> Stone then
    begin
      GoMokuBoard[X, Y] := Stone;
      DrawStone(Canvas, X, Y, False);
      DrawStone(FBuffer.Canvas, X, Y, False);
      if FPlayer1 <> FPlayer2 then
        if ((FPlayerSequence = psPlayer1) and (FPlayer1 = plComputer)) or
           ((FPlayerSequence = psPlayer2) and (FPlayer2 = plComputer)) then
        begin
          Redo;
          Exit;
        end;
      if Assigned(FOnChange) then FOnChange(Self);
      if FiveInARow(True) then
      begin
        if FUseCursors then
          Cursor := crCannotClick
        else
          Cursor := crDefault;
      end else
        if FUseCursors then
          Cursor := crCanClick
        else
          Cursor := crDefault;
      Exit;
    end;
    X := X + 1;
    if X > 19 then
    begin
      X := 1;
      Y := Y + 1;
      if Y > 19 then Exit;
    end;
  end;
end;

{ The Go-Moku Computer Brain }

procedure TGsGoMoku.ComputerSequence;
var
  CanPlay: Boolean;
  OpponentStone, PlayerStone: TGsGoStone;
  I, X, Y: Integer;
  NodeList: TStrings;
  DefensiveNode, OfensiveNode, Node: TGsGoMokuBrainNode;
begin
  CanPlay := ((FPlayerSequence = psPlayer1) and (FPlayer1 = plComputer)) or
    ((FPlayerSequence = psPlayer2) and (FPlayer2 = plComputer));
  if not CanPlay or FiveInARow(False) then Exit;
  if FNodes >= 361 then
  begin
    if Assigned(FOnEndGame) then FOnEndGame(Self, goNone);
    Exit;
  end;

  NodeList := TStringList.Create;

  if FPlayerSequence = psPlayer1 then
  begin
    PlayerStone := goPlayer1;
    OpponentStone := goPlayer2;
  end else
  begin
    PlayerStone := goPlayer2;
    OpponentStone := goPlayer1;
  end;

  if FNodes > 0 then
  begin
    { Defensive }
    if FComputerSkill = csHigh then
      DefensiveTrickCheck(OpponentStone, NodeList);
    HorizontallyCheck(True, OpponentStone, NodeList);
    VerticallyCheck(True, OpponentStone, NodeList);
    ObliqueCheck_TLBR(True, OpponentStone, NodeList);
    ObliqueCheck_TRBL(True, OpponentStone, NodeList);
    { Offensive }
    HorizontallyCheck(False, PlayerStone, NodeList);
    VerticallyCheck(False, PlayerStone, NodeList);
    if FComputerSkill in [csMedium, csHigh] then
    begin
      ObliqueCheck_TLBR(False, PlayerStone, NodeList);
      ObliqueCheck_TRBL(False, PlayerStone, NodeList);
    end;
  end;    
  DefensiveNode := nil;
  OfensiveNode := nil;

  for I := 0 to pred(NodeList.Count) do
    if NodeList.Objects[I] <> nil then
    begin
      Node := TGsGoMokuBrainNode(NodeList.Objects[I]);
      if NodeList[I] = 'Defensive' then
      begin
        if DefensiveNode <> nil then
        begin
          if Node.Level > DefensiveNode.Level then
            DefensiveNode := Node;
        end else
          DefensiveNode := Node;
      end else
      begin
        if OfensiveNode <> nil then
        begin
          if Node.Level > OfensiveNode.Level then
            OfensiveNode := Node;
        end else
          OfensiveNode := Node;
      end;
    end;

  if (OfensiveNode <> nil) and (DefensiveNode <> nil) then
  begin
    if (OfensiveNode.Level = 5) and (DefensiveNode.Level = 5) then
      ChangeStone(OfensiveNode.X, OfensiveNode.Y, OfensiveNode.Stone)
    else
      if (OfensiveNode.Level > DefensiveNode.Level) then
        ChangeStone(OfensiveNode.X, OfensiveNode.Y, OfensiveNode.Stone)
      else
        if (OfensiveNode.Level <= DefensiveNode.Level) then
          ChangeStone(DefensiveNode.X, DefensiveNode.Y, DefensiveNode.Stone);
  end else
    if (OfensiveNode = nil) and (DefensiveNode <> nil) then
    begin
      if DefensiveNode.Level <= 2 then
      begin
        if FComputerSkill in [csMedium, csHigh] then
          ChangeStone(DefensiveNode.X, DefensiveNode.Y, DefensiveNode.Stone)
      end else
        ChangeStone(DefensiveNode.X, DefensiveNode.Y, DefensiveNode.Stone)
    end else
      if (OfensiveNode <> nil) and (DefensiveNode = nil) then
        ChangeStone(OfensiveNode.X, OfensiveNode.Y, OfensiveNode.Stone)
      else begin
        Randomize;
        X := Random(18) + 1;
        Y := Random(18) + 1;
        while GoMokuBoard[X, Y] <> goNone do
        begin
          X := Random(18) + 1;
          Y := Random(18) + 1;
        end;
        ChangeStone(X, Y, PlayerStone);
        FPlayerSequence := SwitchSequence[FPlayerSequence];
      end;

  if (OfensiveNode <> nil) or (DefensiveNode <> nil) then
    FPlayerSequence := SwitchSequence[FPlayerSequence];

  I := pred(NodeList.Count);
  while I > -1 do
  begin
    if NodeList.Objects[I] <> nil then
    begin
      Node := TGsGoMokuBrainNode(NodeList.Objects[I]);
      Node.Free;
    end;
    NodeList.Delete(I);
    I := pred(NodeList.Count);
  end;
  NodeList.Free;

  FNodes := FNodes + 1;
  if Assigned(FOnChange) then FOnChange(Self);
  if FiveInARow(True) then Exit;
  if FNodes >= 361 then
  begin
    if Assigned(FOnEndGame) then FOnEndGame(Self, goNone);
    Exit;
  end;

  if (FPlayerSequence = psPlayer1) and (FPlayer1 = plComputer) then
    ComputerSequence
  else
    if (FPlayerSequence = psPlayer2) and (FPlayer2 = plComputer) then
      ComputerSequence;
end;

procedure TGsGoMoku.CalcBestPosition(Defensive: Boolean; NodeList: TStrings);
begin

end;

procedure AddNode(S: string; Stone: TGsGoStone; X, Y, Level: Integer; NodeList: TStrings);
var
  Node: TGsGoMokuBrainNode;
begin
  if (Level = 0) or not (X in [1..19]) or not(Y in [1..19]) or (S = '') or
    (NodeList = nil) then Exit;
  Node := TGsGoMokuBrainNode.Create;
  Node.X := X;
  Node.Y := Y;
  Node.Level := Level;
  Node.Stone := Stone;
  NodeList.AddObject(S, Node);
end;

function TGsGoMoku.FiveInARow(Draw5InARow: Boolean): Boolean;
var
  X, Y: Integer;
  Stone: TGsGoStone;
begin
  Result := False;
  for Y := 1 to 19 do
  begin
    for X := 1 to 19 do
    begin
      Stone := GoMokuBoard[X, Y];
      if Stone = goNone then Continue;
      
      if (X in [1..15]) then
      begin
        if (GoMokuBoard[X + 1, Y] = Stone) and
           (GoMokuBoard[X + 2, Y] = Stone) and
           (GoMokuBoard[X + 3, Y] = Stone) and
           (GoMokuBoard[X + 4, Y] = Stone) then
        begin
          Result := True;
          if not Draw5InARow then Exit;
          DrawStone(Canvas, X, Y, True);
          DrawStone(Canvas, X + 1, Y, True);
          DrawStone(Canvas, X + 2, Y, True);
          DrawStone(Canvas, X + 3, Y, True);
          DrawStone(Canvas, X + 4, Y, True);
          if Assigned(FOnEndGame) then FOnEndGame(Self, Stone);
          Exit;
        end;
      end;

      if (Y in [1..15]) then
      begin
        if (GoMokuBoard[X, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = Stone) and
           (GoMokuBoard[X, Y + 3] = Stone) and
           (GoMokuBoard[X, Y + 4] = Stone) then
        begin
          Result := True;
          if not Draw5InARow then Exit;
          DrawStone(Canvas, X, Y, True);
          DrawStone(Canvas, X, Y + 1, True);
          DrawStone(Canvas, X, Y + 2, True);
          DrawStone(Canvas, X, Y + 3, True);
          DrawStone(Canvas, X, Y + 4, True);
          if Assigned(FOnEndGame) then FOnEndGame(Self, Stone);
          Exit;
        end;
      end;

      if (X in [1..15]) and (Y in [1..15]) then
      begin
        if (GoMokuBoard[X + 1, Y + 1] = Stone) and
           (GoMokuBoard[X + 2, Y + 2] = Stone) and
           (GoMokuBoard[X + 3, Y + 3] = Stone) and
           (GoMokuBoard[X + 4, Y + 4] = Stone) then
        begin
          Result := True;
          if not Draw5InARow then Exit;
          DrawStone(Canvas, X, Y, True);
          DrawStone(Canvas, X + 1, Y + 1, True);
          DrawStone(Canvas, X + 2, Y + 2, True);
          DrawStone(Canvas, X + 3, Y + 3, True);
          DrawStone(Canvas, X + 4, Y + 4, True);
          if Assigned(FOnEndGame) then FOnEndGame(Self, Stone);
          Exit;
        end;
      end;

      if (X in [5..19]) and (Y in [1..15]) then
      begin
        if (GoMokuBoard[X - 1, Y + 1] = Stone) and
           (GoMokuBoard[X - 2, Y + 2] = Stone) and
           (GoMokuBoard[X - 3, Y + 3] = Stone) and
           (GoMokuBoard[X - 4, Y + 4] = Stone) then
        begin
          Result := True;
          if not Draw5InARow then Exit;
          DrawStone(Canvas, X, Y, True);
          DrawStone(Canvas, X - 1, Y + 1, True);
          DrawStone(Canvas, X - 2, Y + 2, True);
          DrawStone(Canvas, X - 3, Y + 3, True);
          DrawStone(Canvas, X - 4, Y + 4, True);
          if Assigned(FOnEndGame) then FOnEndGame(Self, Stone);
          Exit;
        end;
      end;

    end;
  end;
end;

function TGsGoMoku.DefensiveTrickCheck(Stone: TGsGoStone;
  NodeList: TStrings): Boolean;
var
  X, Y: Integer;
  NodeStone: TGsGoStone;
begin
  Result := False;
  if (Stone = goNone) or (NodeList = nil) then Exit;
  if Stone = goPlayer1 then
    NodeStone := goPlayer2
  else
    NodeStone := goPlayer1;
  for Y := 1 to 19 do
  begin
    for X := 1 to 19 do
    begin
      if (X in [1..15]) and (Y in [1..15]) then
        {
          ???_?
          _XX__
          ???X?
          ???X?
          ???_?
        }
        if (GoMokuBoard[X + 3, Y] = goNone) and
           (GoMokuBoard[X, Y + 1] = goNone) and
           (GoMokuBoard[X + 1, Y + 1] = Stone) and
           (GoMokuBoard[X + 2, Y + 1] = Stone) and
           (GoMokuBoard[X + 3, Y + 1] = goNone) and
           (GoMokuBoard[X + 4, Y + 1] = goNone) and
           (GoMokuBoard[X + 3, Y + 2] = Stone) and
           (GoMokuBoard[X + 3, Y + 3] = Stone) and
           (GoMokuBoard[X + 3, Y + 4] = goNone) then
        begin
          AddNode('Defensive', NodeStone, X + 3, Y + 1, 5, NodeList);
          Result := True;
        end;

      if (X in [1..15]) and (Y in [1..15]) then
        {
          ?_???
          __XX_
          ?X???
          ?X???
          ?_???
        }
        if (GoMokuBoard[X + 1, Y] = goNone) and
           (GoMokuBoard[X, Y + 1] = goNone) and
           (GoMokuBoard[X + 1, Y + 1] = goNone) and
           (GoMokuBoard[X + 2, Y + 1] = Stone) and
           (GoMokuBoard[X + 3, Y + 1] = Stone) and
           (GoMokuBoard[X + 4, Y + 1] = goNone) and
           (GoMokuBoard[X + 1, Y + 2] = Stone) and
           (GoMokuBoard[X + 1, Y + 3] = Stone) and
           (GoMokuBoard[X + 1, Y + 4] = goNone) then
        begin
          AddNode('Defensive', NodeStone, X + 1, Y + 1, 5, NodeList);
          Result := True;
        end;

      if (X in [1..15]) and (Y in [1..15]) then
        {
          ?_???
          ?X???
          ?X???
          __XX_
          ?_???
        }
        if (GoMokuBoard[X + 1, Y] = goNone) and
           (GoMokuBoard[X + 1, Y + 1] = Stone) and
           (GoMokuBoard[X + 1, Y + 2] = Stone) and
           (GoMokuBoard[X, Y + 3] = goNone) and
           (GoMokuBoard[X + 1, Y + 3] = goNone) and
           (GoMokuBoard[X + 2, Y + 3] = Stone) and
           (GoMokuBoard[X + 3, Y + 3] = Stone) and
           (GoMokuBoard[X + 4, Y + 3] = goNone) and
           (GoMokuBoard[X + 1, Y + 4] = goNone) then
        begin
          AddNode('Defensive', NodeStone, X + 1, Y + 3, 5, NodeList);
          Result := True;
        end;

      if (X in [1..15]) and (Y in [1..15]) then
        {
          ???_?
          ???X?
          ???X?
          _XX__
          ???_?
        }
        if (GoMokuBoard[X + 3, Y] = goNone) and
           (GoMokuBoard[X + 3, Y + 1] = Stone) and
           (GoMokuBoard[X + 3, Y + 2] = Stone) and
           (GoMokuBoard[X, Y + 3] = goNone) and
           (GoMokuBoard[X + 1, Y + 3] = Stone) and
           (GoMokuBoard[X + 2, Y + 3] = Stone) and
           (GoMokuBoard[X + 3, Y + 3] = goNone) and
           (GoMokuBoard[X + 4, Y + 3] = goNone) and
           (GoMokuBoard[X + 3, Y + 4] = goNone) then
        begin
          AddNode('Defensive', NodeStone, X + 3, Y + 3, 5, NodeList);
          Result := True;
        end;

      if (X in [1..15]) and (Y in [1..15]) then
        {
          ??_??
          ??X??
          _X_X_
          ??X??
          ??_??
        }
        if (GoMokuBoard[X + 2, Y] = goNone) and
           (GoMokuBoard[X + 2, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = goNone) and
           (GoMokuBoard[X + 1, Y + 2] = Stone) and
           (GoMokuBoard[X + 2, Y + 2] = goNone) and
           (GoMokuBoard[X + 3, Y + 2] = Stone) and
           (GoMokuBoard[X + 4, Y + 2] = goNone) and
           (GoMokuBoard[X + 2, Y + 3] = Stone) and
           (GoMokuBoard[X + 2, Y + 4] = goNone) then
        begin
          AddNode('Defensive', NodeStone, X + 2, Y + 2, 5, NodeList);
          Result := True;
        end;

      if (X in [1..15]) and (Y in [1..15]) then
        {
          _???_
          ?X?X?
          ??_??
          ?X?X?
          _???_
        }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 4, Y] = goNone) and
           (GoMokuBoard[X + 1, Y + 1] = Stone) and
           (GoMokuBoard[X + 3, Y + 1] = Stone) and
           (GoMokuBoard[X + 2, Y + 2] = goNone) and
           (GoMokuBoard[X + 1, Y + 3] = Stone) and
           (GoMokuBoard[X + 3, Y + 3] = Stone) and
           (GoMokuBoard[X, Y + 4] = goNone) and
           (GoMokuBoard[X + 4, Y + 4] = goNone) then
        begin
          AddNode('Defensive', NodeStone, X + 2, Y + 2, 5, NodeList);
          Result := True;
        end;

    end;
  end;
end;

function TGsGoMoku.HorizontallyCheck(Defensive: Boolean; Stone: TGsGoStone;
  NodeList: TStrings): Boolean;
var
  X, Y: Integer;
  NodeStone, OppStone: TGsGoStone;
  ItemStr: string;
begin
  Result := False;
  if (Stone = goNone) or (NodeList = nil) then Exit;

  if Defensive then
  begin
    ItemStr := 'Defensive';
    if Stone = goPlayer1 then
    begin
      NodeStone := goPlayer2;
      OppStone := goPlayer1;
    end else
    begin
      NodeStone := goPlayer1;
      OppStone := goPlayer2;
    end;
  end else
  begin
    ItemStr := 'Ofensive';
    NodeStone := Stone;
    if Stone = goPlayer1 then
      OppStone := goPlayer2
    else
      OppStone := goPlayer1;
  end;

  for Y := 1 to 19 do
  begin
    for X := 1 to 19 do
    begin

      if X <= 14 then
      begin
        { _XX_X_ }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y] = Stone) and
           (GoMokuBoard[X + 2, Y] = Stone) and
           (GoMokuBoard[X + 3, Y] = goNone) and
           (GoMokuBoard[X + 4, Y] = Stone) and
           (GoMokuBoard[X + 5, Y] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X + 3, Y, 4, NodeList);
          Result := True;
        end;
        { _X_XX_ }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y] = Stone) and
           (GoMokuBoard[X + 2, Y] = goNone) and
           (GoMokuBoard[X + 3, Y] = Stone) and
           (GoMokuBoard[X + 4, Y] = Stone) and
           (GoMokuBoard[X + 5, Y] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X + 2, Y, 4, NodeList);
          Result := True;
        end;
      end;

      if X <= 15 then
      begin
        { XXXX_ }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X + 1, Y] = Stone) and
           (GoMokuBoard[X + 2, Y] = Stone) and
           (GoMokuBoard[X + 3, Y] = Stone) and
           (GoMokuBoard[X + 4, Y] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X + 4, Y, 5, NodeList);
          Result := True;
        end;
        { XXX_X }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X + 1, Y] = Stone) and
           (GoMokuBoard[X + 2, Y] = Stone) and
           (GoMokuBoard[X + 3, Y] = goNone) and
           (GoMokuBoard[X + 4, Y] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X + 3, Y, 5, NodeList);
          Result := True;
        end;
        { XX_XX }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X + 1, Y] = Stone) and
           (GoMokuBoard[X + 2, Y] = goNone) and
           (GoMokuBoard[X + 3, Y] = Stone) and
           (GoMokuBoard[X + 4, Y] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X + 2, Y, 5, NodeList);
          Result := True;
        end;
        { X_XXX }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X + 1, Y] = goNone) and
           (GoMokuBoard[X + 2, Y] = Stone) and
           (GoMokuBoard[X + 3, Y] = Stone) and
           (GoMokuBoard[X + 4, Y] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X + 1, Y, 5, NodeList);
          Result := True;
        end;
        { _XXXX }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y] = Stone) and
           (GoMokuBoard[X + 2, Y] = Stone) and
           (GoMokuBoard[X + 3, Y] = Stone) and
           (GoMokuBoard[X + 4, Y] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 5, NodeList);
          Result := True;
        end;
        { _XXX_ }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y] = Stone) and
           (GoMokuBoard[X + 2, Y] = Stone) and
           (GoMokuBoard[X + 3, Y] = Stone) and
           (GoMokuBoard[X + 4, Y] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 3, NodeList);
          AddNode(ItemStr, NodeStone, X + 4, Y, 3, NodeList);
          Result := True;
        end;
        { _XX_X }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y] = Stone) and
           (GoMokuBoard[X + 2, Y] = Stone) and
           (GoMokuBoard[X + 3, Y] = goNone) and
           (GoMokuBoard[X + 4, Y] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 3, NodeList);
          AddNode(ItemStr, NodeStone, X + 3, Y, 3, NodeList);
          Result := True;
        end;
        { _X_XX }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y] = Stone) and
           (GoMokuBoard[X + 2, Y] = goNone) and
           (GoMokuBoard[X + 3, Y] = Stone) and
           (GoMokuBoard[X + 4, Y] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 3, NodeList);
          AddNode(ItemStr, NodeStone, X + 2, Y, 3, NodeList);
          Result := True;
        end;
        { _XXXO }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y] = Stone) and
           (GoMokuBoard[X + 2, Y] = Stone) and
           (GoMokuBoard[X + 3, Y] = Stone) and
           (GoMokuBoard[X + 4, Y] = OppStone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 2, NodeList);
          Result := True;
        end;
        { OXXX_ }
        if (GoMokuBoard[X, Y] = OppStone) and
           (GoMokuBoard[X + 1, Y] = Stone) and
           (GoMokuBoard[X + 2, Y] = Stone) and
           (GoMokuBoard[X + 3, Y] = Stone) and
           (GoMokuBoard[X + 4, Y] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X + 4, Y, 2, NodeList);
          Result := True;
        end;
      end;

      if X <= 16 then
      begin
        { _XX_ }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y] = Stone) and
           (GoMokuBoard[X + 2, Y] = Stone) and
           (GoMokuBoard[X + 3, Y] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 2, NodeList);
          AddNode(ItemStr, NodeStone, X + 3, Y, 2, NodeList);
          Result := True;
        end;
        { _XXO }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y] = Stone) and
           (GoMokuBoard[X + 2, Y] = Stone) and
           (GoMokuBoard[X + 3, Y] = OppStone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 1, NodeList);
          Result := True;
        end;
        { OXX_ }
        if (GoMokuBoard[X, Y] = OppStone) and
           (GoMokuBoard[X + 1, Y] = Stone) and
           (GoMokuBoard[X + 2, Y] = Stone) and
           (GoMokuBoard[X + 3, Y] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X + 3, Y, 1, NodeList);
          Result := True;
        end;
      end;

      if not Defensive and (X <= 18) then
      begin
        { _O }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y] <> goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 1, NodeList);
          Result := True;
        end;
        { O_ }
        if (GoMokuBoard[X, Y] <> goNone) and
           (GoMokuBoard[X + 1, Y] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X + 1, Y, 1, NodeList);
          Result := True;
        end;
      end;

    end;
  end;
end;

function TGsGoMoku.VerticallyCheck(Defensive: Boolean; Stone: TGsGoStone;
  NodeList: TStrings): Boolean;
var
  X, Y: Integer;
  NodeStone, OppStone: TGsGoStone;
  ItemStr: string;
begin
  Result := False;
  if (Stone = goNone) or (NodeList = nil) then Exit;

  if Defensive then
  begin
    ItemStr := 'Defensive';
    if Stone = goPlayer1 then
    begin
      NodeStone := goPlayer2;
      OppStone := goPlayer1;
    end else
    begin
      NodeStone := goPlayer1;
      OppStone := goPlayer2;
    end;
  end else
  begin
    ItemStr := 'Ofensive';
    NodeStone := Stone;
    if Stone = goPlayer1 then
      OppStone := goPlayer2
    else
      OppStone := goPlayer1;
  end;

  for X := 1 to 19 do
  begin
    for Y := 1 to 19 do
    begin

      if Y <= 14 then
      begin
        { _XX_X_ }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = Stone) and
           (GoMokuBoard[X, Y + 3] = goNone) and
           (GoMokuBoard[X, Y + 4] = Stone) and
           (GoMokuBoard[X, Y + 5] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y + 3, 4, NodeList);
          Result := True;
        end;
        { _X_XX_ }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = goNone) and
           (GoMokuBoard[X, Y + 3] = Stone) and
           (GoMokuBoard[X, Y + 4] = Stone) and
           (GoMokuBoard[X, Y + 5] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y + 2, 4, NodeList);
          Result := True;
        end;
      end;

      if Y <= 15 then
      begin
        { XXXX_ }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = Stone) and
           (GoMokuBoard[X, Y + 3] = Stone) and
           (GoMokuBoard[X, Y + 4] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y + 4, 5, NodeList);
          Result := True;
        end;
        { XXX_X }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = Stone) and
           (GoMokuBoard[X, Y + 3] = goNone) and
           (GoMokuBoard[X, Y + 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y + 3, 5, NodeList);
          Result := True;
        end;
        { XX_XX }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = goNone) and
           (GoMokuBoard[X, Y + 3] = Stone) and
           (GoMokuBoard[X, Y + 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y + 2, 5, NodeList);
          Result := True;
        end;
        { X_XXX }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X, Y + 1] = goNone) and
           (GoMokuBoard[X, Y + 2] = Stone) and
           (GoMokuBoard[X, Y + 3] = Stone) and
           (GoMokuBoard[X, Y + 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y + 1, 5, NodeList);
          Result := True;
        end;
        { _XXXX }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = Stone) and
           (GoMokuBoard[X, Y + 3] = Stone) and
           (GoMokuBoard[X, Y + 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 5, NodeList);
          Result := True;
        end;
        { _XXX_ }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = Stone) and
           (GoMokuBoard[X, Y + 3] = Stone) and
           (GoMokuBoard[X, Y + 4] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 3, NodeList);
          AddNode(ItemStr, NodeStone, X, Y + 4, 3, NodeList);
          Result := True;
        end;
        { _XX_X }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = Stone) and
           (GoMokuBoard[X, Y + 3] = goNone) and
           (GoMokuBoard[X, Y + 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 3, NodeList);
          AddNode(ItemStr, NodeStone, X, Y + 3, 3, NodeList);
          Result := True;
        end;
        { _X_XX }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = goNone) and
           (GoMokuBoard[X, Y + 3] = Stone) and
           (GoMokuBoard[X, Y + 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 3, NodeList);
          AddNode(ItemStr, NodeStone, X, Y + 2, 3, NodeList);
          Result := True;
        end;
        { _XXXO }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = Stone) and
           (GoMokuBoard[X, Y + 3] = Stone) and
           (GoMokuBoard[X, Y + 4] = OppStone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 2, NodeList);
          Result := True;
        end;
        { OXXX_ }
        if (GoMokuBoard[X, Y] = OppStone) and
           (GoMokuBoard[X, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = Stone) and
           (GoMokuBoard[X, Y + 3] = Stone) and
           (GoMokuBoard[X, Y + 4] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y + 4, 2, NodeList);
          Result := True;
        end;
      end;

      if Y <= 16 then
      begin
        { _XX_ }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = Stone) and
           (GoMokuBoard[X, Y + 3] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 2, NodeList);
          AddNode(ItemStr, NodeStone, X, Y + 3, 2, NodeList);
          Result := True;
        end;
        { _XXO }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = Stone) and
           (GoMokuBoard[X, Y + 3] = OppStone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 1, NodeList);
          Result := True;
        end;
        { OXX_ }
        if (GoMokuBoard[X, Y] = OppStone) and
           (GoMokuBoard[X, Y + 1] = Stone) and
           (GoMokuBoard[X, Y + 2] = Stone) and
           (GoMokuBoard[X, Y + 3] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y + 3, 1, NodeList);
          Result := True;
        end;
      end;

      if not Defensive and (X <= 18) then
      begin
        { _O }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X, Y + 1] <> goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 1, NodeList);
          Result := True;
        end;
        { O_ }
        if (GoMokuBoard[X, Y] = OppStone) and
           (GoMokuBoard[X, Y + 1] <> goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y + 1, 1, NodeList);
          Result := True;
        end;
      end;

    end;
  end;
end;

function TGsGoMoku.ObliqueCheck_TLBR(Defensive: Boolean; Stone: TGsGoStone;
  NodeList: TStrings): Boolean;
var
  X, Y: Integer;
  NodeStone, OppStone: TGsGoStone;
  ItemStr: string;
begin
  Result := False;
  if (Stone = goNone) or (NodeList = nil) then Exit;

  if Defensive then
  begin
    ItemStr := 'Defensive';
    if Stone = goPlayer1 then
    begin
      NodeStone := goPlayer2;
      OppStone := goPlayer1;
    end else
    begin
      NodeStone := goPlayer1;
      OppStone := goPlayer2;
    end;
  end else
  begin
    ItemStr := 'Ofensive';
    NodeStone := Stone;
    if Stone = goPlayer1 then
      OppStone := goPlayer2
    else
      OppStone := goPlayer1;
  end;

  for Y := 1 to 19 do
  begin
    for X := 1 to 19 do
    begin

      if (X in [6..19]) and (Y in [6..19]) then
      begin
        { _XX_X_ }
        if (GoMokuBoard[X - 5, Y - 5] = goNone) and
           (GoMokuBoard[X - 4, Y - 4] = Stone) and
           (GoMokuBoard[X - 3, Y - 3] = Stone) and
           (GoMokuBoard[X - 2, Y - 2] = goNone) and
           (GoMokuBoard[X - 1, Y - 1] = Stone) and
           (GoMokuBoard[X, Y] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X - 2, Y - 2, 4, NodeList);
          Result := True;
        end;
        { _X_XX_ }
        if (GoMokuBoard[X - 5, Y - 5] = goNone) and
           (GoMokuBoard[X - 4, Y - 4] = Stone) and
           (GoMokuBoard[X - 3, Y - 3] = goNone) and
           (GoMokuBoard[X - 2, Y - 2] = Stone) and
           (GoMokuBoard[X - 1, Y - 1] = Stone) and
           (GoMokuBoard[X, Y] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X - 3, Y - 3, 4, NodeList);
          Result := True;
        end;
      end;

      if (X in [5..19]) and (Y in [5..19]) then
      begin
        { XXXX_ }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X - 1, Y - 1] = Stone) and
           (GoMokuBoard[X - 2, Y - 2] = Stone) and
           (GoMokuBoard[X - 3, Y - 3] = Stone) and
           (GoMokuBoard[X - 4, Y - 4] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X - 4, Y - 4, 5, NodeList);
          Result := True;
        end;
        { XXX_X }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X - 1, Y - 1] = Stone) and
           (GoMokuBoard[X - 2, Y - 2] = Stone) and
           (GoMokuBoard[X - 3, Y - 3] = goNone) and
           (GoMokuBoard[X - 4, Y - 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X - 3, Y - 3, 5, NodeList);
          Result := True;
        end;
        { XX_XX }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X - 1, Y - 1] = Stone) and
           (GoMokuBoard[X - 2, Y - 2] = goNone) and
           (GoMokuBoard[X - 3, Y - 3] = Stone) and
           (GoMokuBoard[X - 4, Y - 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X - 2, Y - 2, 5, NodeList);
          Result := True;
        end;
        { X_XXX }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X - 1, Y - 1] = goNone) and
           (GoMokuBoard[X - 2, Y - 2] = Stone) and
           (GoMokuBoard[X - 3, Y - 3] = Stone) and
           (GoMokuBoard[X - 4, Y - 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X - 1, Y - 1, 5, NodeList);
          Result := True;
        end;
        { _XXXX }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X - 1, Y - 1] = Stone) and
           (GoMokuBoard[X - 2, Y - 2] = Stone) and
           (GoMokuBoard[X - 3, Y - 3] = Stone) and
           (GoMokuBoard[X - 4, Y - 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 5, NodeList);
          Result := True;
        end;
        { _XXX_ }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X - 1, Y - 1] = Stone) and
           (GoMokuBoard[X - 2, Y - 2] = Stone) and
           (GoMokuBoard[X - 3, Y - 3] = Stone) and
           (GoMokuBoard[X - 4, Y - 4] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 3, NodeList);
          AddNode(ItemStr, NodeStone, X - 4, Y - 4, 3, NodeList);
          Result := True;
        end;
        { _XX_X }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X - 1, Y - 1] = Stone) and
           (GoMokuBoard[X - 2, Y - 2] = Stone) and
           (GoMokuBoard[X - 3, Y - 3] = goNone) and
           (GoMokuBoard[X - 4, Y - 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 3, NodeList);
          AddNode(ItemStr, NodeStone, X - 3, Y - 3, 3, NodeList);
          Result := True;
        end;
        { _X_XX }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X - 1, Y - 1] = Stone) and
           (GoMokuBoard[X - 2, Y - 2] = goNone) and
           (GoMokuBoard[X - 3, Y - 3] = Stone) and
           (GoMokuBoard[X - 4, Y - 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 3, NodeList);
          AddNode(ItemStr, NodeStone, X - 2, Y - 2, 3, NodeList);
          Result := True;
        end;
        { _XXXO }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X - 1, Y - 1] = Stone) and
           (GoMokuBoard[X - 1, Y - 2] = Stone) and
           (GoMokuBoard[X - 1, Y - 3] = Stone) and
           (GoMokuBoard[X - 1, Y - 4] = OppStone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 2, NodeList);
          Result := True;
        end;
        { OXXX_ }
        if (GoMokuBoard[X, Y] = OppStone) and
           (GoMokuBoard[X - 1, Y - 1] = Stone) and
           (GoMokuBoard[X - 2, Y - 2] = Stone) and
           (GoMokuBoard[X - 3, Y - 3] = Stone) and
           (GoMokuBoard[X - 4, Y - 4] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X - 4, Y - 4, 2, NodeList);
          Result := True;
        end;
      end;

      if (X in [4..19]) and (Y in [4..19]) then
      begin
        { _XX_ }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X - 1, Y - 1] = Stone) and
           (GoMokuBoard[X - 2, Y - 2] = Stone) and
           (GoMokuBoard[X - 3, Y - 3] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 2, NodeList);
          AddNode(ItemStr, NodeStone, X - 3, Y - 3, 2, NodeList);
          Result := True;
        end;
        { _XXO }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X - 1, Y - 1] = Stone) and
           (GoMokuBoard[X - 2, Y - 2] = Stone) and
           (GoMokuBoard[X - 3, Y - 3] = OppStone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 1, NodeList);
          Result := True;
        end;
        { OXX_ }
        if (GoMokuBoard[X, Y] = OppStone) and
           (GoMokuBoard[X - 1, Y - 1] = Stone) and
           (GoMokuBoard[X - 2, Y - 2] = Stone) and
           (GoMokuBoard[X - 3, Y - 3] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X - 3, Y - 3, 1, NodeList);
          Result := True;
        end;
      end;

      if not Defensive and (X <= 18) then
      begin
        { _O }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X - 1, Y - 1] <> goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 1, NodeList);
          Result := True;
        end;
        { O_ }
        if (GoMokuBoard[X, Y] <> goNone) and
           (GoMokuBoard[X - 1, Y - 1] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X - 1, Y - 1, 1, NodeList);
          Result := True;
        end;
      end;

    end;
  end;
end;

function TGsGoMoku.ObliqueCheck_TRBL(Defensive: Boolean; Stone: TGsGoStone;
  NodeList: TStrings): Boolean;
var
  X, Y: Integer;
  NodeStone, OppStone: TGsGoStone;
  ItemStr: string;
begin
  Result := False;
  if (Stone = goNone) or (NodeList = nil) then Exit;

  if Defensive then
  begin
    ItemStr := 'Defensive';
    if Stone = goPlayer1 then
    begin
      NodeStone := goPlayer2;
      OppStone := goPlayer1;
    end else
    begin
      NodeStone := goPlayer1;
      OppStone := goPlayer2;
    end;
  end else
  begin
    ItemStr := 'Ofensive';
    NodeStone := Stone;
    if Stone = goPlayer1 then
      OppStone := goPlayer2
    else
      OppStone := goPlayer1;
  end;

  for Y := 1 to 19 do
  begin
    for X := 1 to 19 do
    begin

      if (X in [1..14]) and (Y in [6..19]) then
      begin
        { _XX_X_ }
        if (GoMokuBoard[X + 5, Y - 5] = goNone) and
           (GoMokuBoard[X + 4, Y - 4] = Stone) and
           (GoMokuBoard[X + 3, Y - 3] = Stone) and
           (GoMokuBoard[X + 2, Y - 2] = goNone) and
           (GoMokuBoard[X + 1, Y - 1] = Stone) and
           (GoMokuBoard[X, Y] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X + 2, Y - 2, 4, NodeList);
          Result := True;
        end;
        { _X_XX_ }
        if (GoMokuBoard[X + 5, Y - 5] = goNone) and
           (GoMokuBoard[X + 4, Y - 4] = Stone) and
           (GoMokuBoard[X + 3, Y - 3] = goNone) and
           (GoMokuBoard[X + 2, Y - 2] = Stone) and
           (GoMokuBoard[X + 1, Y - 1] = Stone) and
           (GoMokuBoard[X, Y] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X + 3, Y - 3, 4, NodeList);
          Result := True;
        end;
      end;

      if (X in [1..15]) and (Y in [5..19]) then
      begin
        { XXXX_ }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X + 1, Y - 1] = Stone) and
           (GoMokuBoard[X + 2, Y - 2] = Stone) and
           (GoMokuBoard[X + 3, Y - 3] = Stone) and
           (GoMokuBoard[X + 4, Y - 4] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X + 4, Y - 4, 5, NodeList);
          Result := True;
        end;
        { XXX_X }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X + 1, Y - 1] = Stone) and
           (GoMokuBoard[X + 2, Y - 2] = Stone) and
           (GoMokuBoard[X + 3, Y - 3] = goNone) and
           (GoMokuBoard[X + 4, Y - 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X + 3, Y - 3, 5, NodeList);
          Result := True;
        end;
        { XX_XX }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X + 1, Y - 1] = Stone) and
           (GoMokuBoard[X + 2, Y - 2] = goNone) and
           (GoMokuBoard[X + 3, Y - 3] = Stone) and
           (GoMokuBoard[X + 4, Y - 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X + 2, Y - 2, 5, NodeList);
          Result := True;
        end;
        { X_XXX }
        if (GoMokuBoard[X, Y] = Stone) and
           (GoMokuBoard[X + 1, Y - 1] = goNone) and
           (GoMokuBoard[X + 2, Y - 2] = Stone) and
           (GoMokuBoard[X + 3, Y - 3] = Stone) and
           (GoMokuBoard[X + 4, Y - 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X + 1, Y - 1, 5, NodeList);
          Result := True;
        end;
        { _XXXX }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y - 1] = Stone) and
           (GoMokuBoard[X + 2, Y - 2] = Stone) and
           (GoMokuBoard[X + 3, Y - 3] = Stone) and
           (GoMokuBoard[X + 4, Y - 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 5, NodeList);
          Result := True;
        end;
        { _XXX_ }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y - 1] = Stone) and
           (GoMokuBoard[X + 2, Y - 2] = Stone) and
           (GoMokuBoard[X + 3, Y - 3] = Stone) and
           (GoMokuBoard[X + 4, Y - 4] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 3, NodeList);
          AddNode(ItemStr, NodeStone, X + 4, Y - 4, 3, NodeList);
          Result := True;
        end;
        { _XX_X }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y - 1] = Stone) and
           (GoMokuBoard[X + 2, Y - 2] = Stone) and
           (GoMokuBoard[X + 3, Y - 3] = goNone) and
           (GoMokuBoard[X + 4, Y - 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 3, NodeList);
          AddNode(ItemStr, NodeStone, X + 3, Y - 3, 3, NodeList);
          Result := True;
        end;
        { _X_XX }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y - 1] = Stone) and
           (GoMokuBoard[X + 2, Y - 2] = goNone) and
           (GoMokuBoard[X + 3, Y - 3] = Stone) and
           (GoMokuBoard[X + 4, Y - 4] = Stone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 3, NodeList);
          AddNode(ItemStr, NodeStone, X + 2, Y - 2, 3, NodeList);
          Result := True;
        end;
        { _XXXO }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y - 1] = Stone) and
           (GoMokuBoard[X + 1, Y - 2] = Stone) and
           (GoMokuBoard[X + 1, Y - 3] = Stone) and
           (GoMokuBoard[X + 1, Y - 4] = OppStone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 2, NodeList);
          Result := True;
        end;
        { OXXX_ }
        if (GoMokuBoard[X, Y] = OppStone) and
           (GoMokuBoard[X + 1, Y - 1] = Stone) and
           (GoMokuBoard[X + 2, Y - 2] = Stone) and
           (GoMokuBoard[X + 3, Y - 3] = Stone) and
           (GoMokuBoard[X + 4, Y - 4] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X + 4, Y - 4, 2, NodeList);
          Result := True;
        end;
      end;

      if (X in [1..16]) and (Y in [4..19]) then
      begin
        { _XX_ }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y - 1] = Stone) and
           (GoMokuBoard[X + 2, Y - 2] = Stone) and
           (GoMokuBoard[X + 3, Y - 3] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 2, NodeList);
          AddNode(ItemStr, NodeStone, X + 3, Y - 3, 2, NodeList);
          Result := True;
        end;
        { _XXO }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y - 1] = Stone) and
           (GoMokuBoard[X + 2, Y - 2] = Stone) and
           (GoMokuBoard[X + 3, Y - 3] = OppStone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 1, NodeList);
          Result := True;
        end;
        { OXX_ }
        if (GoMokuBoard[X, Y] = OppStone) and
           (GoMokuBoard[X + 1, Y - 1] = Stone) and
           (GoMokuBoard[X + 2, Y - 2] = Stone) and
           (GoMokuBoard[X + 3, Y - 3] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X + 3, Y - 3, 1, NodeList);
          Result := True;
        end;
      end;

      if not Defensive and (X <= 18) then
      begin
        { _O }
        if (GoMokuBoard[X, Y] = goNone) and
           (GoMokuBoard[X + 1, Y - 1] <> goNone) then
        begin
          AddNode(ItemStr, NodeStone, X, Y, 1, NodeList);
          Result := True;
        end;
        { O_ }
        if (GoMokuBoard[X, Y] <> goNone) and
           (GoMokuBoard[X + 1, Y - 1] = goNone) then
        begin
          AddNode(ItemStr, NodeStone, X + 1, Y - 1, 1, NodeList);
          Result := True;
        end;
      end;

    end;
  end;
end;

procedure Register;
begin
  RegisterComponents('G''s Games', [TGsGoMoku]);
end;

end.
